-- This module provides the type of natural numbers and some of its
-- properties.

open import Equality
open import Decidable

module Nat where

-- ----------------------------------------------------------------------
-- Definition

-- The type of natural numbers.
data Nat : Set where
  zero : Nat
  succ : Nat -> Nat

infixl 6 _+_ _-_
infixl 7 _*_

-- Addition.
_+_ : Nat -> Nat -> Nat
zero + m = m
succ n + m = succ (n + m)

-- Truncated subtraction.
_-_ : Nat -> Nat -> Nat
n - zero = n
zero - succ m = zero
succ n - succ m = n - m

-- Multiplication.
_*_ : Nat -> Nat -> Nat
zero * m = zero
succ n * m = m + n * m

-- The following directive enables constants such as 3 to be used
-- instead of succ (succ (succ zero)).
{-# BUILTIN NATURAL Nat #-}

-- ----------------------------------------------------------------------
-- Operations

infix 5 _<=_

-- The less-than-or-equal relation.
data _<=_ : Nat -> Nat -> Set where
  leq-zero : ∀ {m} -> 0 <= m
  leq-succ : ∀ {n m} -> n <= m -> succ n <= succ m

-- ----------------------------------------------------------------------
-- Properties

-- Successor is an injective function.
lemma-succ-cancellation : ∀ {n m} -> succ n == succ m -> n == m
lemma-succ-cancellation refl = refl

-- Successor is order-reflecting.
lemma-leq-cancellation : ∀ {n m} -> succ n <= succ m -> n <= m
lemma-leq-cancellation (leq-succ x) = x

-- ----------------------------------------------------------------------
-- Decidable properties.

infix 8 _==?_ _<=?_

-- Equality is decidable.
_==?_ : (n m : Nat) -> Decidable (n == m)
zero ==? zero = yes refl
zero ==? succ m = no (λ ())
succ n ==? zero = no (λ ())
succ n ==? succ m with n ==? m
succ n ==? succ m | yes w = yes (context succ w)
succ n ==? succ m | no w = no (λ k → w (lemma-succ-cancellation k))

-- Leq is decidable.
_<=?_ : (n m : Nat) -> Decidable (n <= m)
zero <=? m = yes leq-zero
succ n <=? zero = no (λ ())
succ n <=? succ m with n <=? m
succ n <=? succ m | yes w = yes (leq-succ w)
succ n <=? succ m | no w = no (λ k → w (lemma-leq-cancellation k))
