-- This file provides two versions of the Reidemeister-Schreier
-- theorem for monoids.
--
-- * The simplified Reidemeister-Schreier theorem assumes that X and Y
--   are sets of generators for the same monoid. In this case, there
--   are translations from X to Y* and from Y to X*, and no cosets are
--   required.
--
-- * The full Reidemeister-Schreier theorem is Theorem 4.2 from the
--   accompanying paper. It assumes that Y generates some monoid and X
--   generates a submonoid, as described in the paper.

open import Proposition
open import Equality as Eq using (_==_)
open import Inspect

open import Word
open import Word-Lemmas

module Reidemeister-Schreier where

-- ----------------------------------------------------------------------
-- * The simplified Reidemeister-Schreier theorem for monoids

-- We assume that X and Y are sets of generators, Γ and Δ are sets of
-- relations over X and Y, respectively, and there are translations
-- f : X -> Y* and g : Y -> X* compatible with Γ and Δ. In this case,
-- it follows that ⟨X | Γ⟩ and ⟨Y | Δ⟩ are isomorphic monoids.

module Reidemeister-Schreier-Simplified
       {X : Set} {Γ : Context X} 
       {Y : Set} {Δ : Context Y} 
       (f : X -> Word Y)
       (g : Y -> Word X)
       (hypA : ∀ (x : X) -> Γ ⊢ gen x === (g *) (f x))
       (hypB : ∀ (u t : Word Y) -> u === t ∈ Δ -> Γ ⊢ (g *) u === (g *) t)
  where

  open Monoid-Equational

  lemma-a : ∀ (w : Word X) -> Γ ⊢ w === (g *) ((f *) w)
  lemma-a (gen x) = hypA x
  lemma-a ε = refl
  lemma-a (w • v) = cong (lemma-a w) (lemma-a v)

  lemma-b : ∀ {u t : Word Y} -> Δ ⊢ u === t -> Γ ⊢ (g *) u === (g *) t
  lemma-b (axiom x) = hypB _ _ x
  lemma-b refl = refl
  lemma-b (symm hyp) = lemma-b hyp reversed
  lemma-b (trans hyp1 hyp2) = trans (lemma-b hyp1) (lemma-b hyp2)
  lemma-b (cong hyp1 hyp2) = cong (lemma-b hyp1) (lemma-b hyp2)
  lemma-b assoc = assoc
  lemma-b left-unit = left-unit
  lemma-b right-unit = right-unit

  reidemeister-schreier-simplified : (w v : Word X) -> Δ ⊢ (f *) w === (f *) v -> Γ ⊢ w === v
  reidemeister-schreier-simplified w v hyp = 
    equational w
            by lemma-a w
        equals (g *) ((f *) w)
            by lemma-b hyp
        equals (g *) ((f *) v)
            by lemma-a v reversed
        equals v

-- ----------------------------------------------------------------------
-- * The full Reidemeister-Schreier theorem for monoids

-- We start with some notations.

-- Extend a function from generators to words, with accumulator.
_** : {Y X C : Set} -> (C -> Y -> Word X × C) -> (C -> Word Y -> Word X × C)
_** h c (gen y) = h c y
_** h c ε = (ε , c)
_** h c (w • u) with (h **) c w
_** h c (w • u) | (w' , c') with (h **) c' u
_** h c (w • u) | (w' , c') | (u' , c'') = (w' • u' , c'')

-- Abbreviation: We write Γ ⊢ₚ (u , c) === (t , c')
-- for Γ ⊢ u === t and c == c'.
_⊢ₚ_===_ : {X C : Set} (Γ : Context X) -> Word X × C -> Word X × C -> Set
Γ ⊢ₚ (u , c) === (t , c') = Γ ⊢ u === t ∧ c == c'

-- Reflexivity, symmetry, and transitivity for ⊢ₚ.
lemma-refl-⊢ₚ : ∀ {X Γ C} {p q : Word X × C} -> p == q -> Γ ⊢ₚ p === q
lemma-refl-⊢ₚ {p = u1 , c1} {u2 , c2} Eq.refl = (refl , Eq.refl)

lemma-symm-⊢ₚ : ∀ {X Γ C} {p q : Word X × C} -> Γ ⊢ₚ p === q -> Γ ⊢ₚ q === p
lemma-symm-⊢ₚ {p = u1 , c1} {u2 , c2} (hyp1 , eq1) = (symm hyp1 , Eq.symm eq1)

lemma-trans-⊢ₚ : ∀ {X Γ C} {p q r : Word X × C} -> Γ ⊢ₚ p === q -> Γ ⊢ₚ q === r -> Γ ⊢ₚ p === r
lemma-trans-⊢ₚ {p = u1 , c1} {u2 , c2} {u3 , c3} (hyp1 , eq1) (hyp2 , eq2) = (trans hyp1 hyp2 , Eq.trans eq1 eq2)

-- Two obvious lemmas about ⊢ₚ.
lemma-⊢ₚ-fst : ∀ {X Γ C} (lhs rhs : Word X × C) -> Γ ⊢ₚ lhs === rhs -> Γ ⊢ fst lhs === fst rhs
lemma-⊢ₚ-fst (w , c) (w' , c') hyp = fst hyp

lemma-⊢ₚ-snd : ∀ {X Γ C} (lhs rhs : Word X × C) -> Γ ⊢ₚ lhs === rhs -> snd lhs == snd rhs
lemma-⊢ₚ-snd (w , c) (w' , c') hyp = snd hyp

-- The Reidemeister-Schreier theorem. We assume that X and Y are sets
-- of generators, Γ and Δ are sets of relations over X and Y,
-- respectively, C is a set (typically a set of coset representatives,
-- although this is not assumed), and there are translations
-- f : X -> Y* and h : C × Y -> X* × C, subject to two assumptions.
-- In that case, Γ is a complete set of relations for the submonoid of
-- Y* generated by X.
module Reidemeister-Schreier-Full
       {X : Set} {Γ : Context X}
       {Y : Set} {Δ : Context Y}
       (C : Set)
       (I : C)
       (f : X -> Word Y)
       (h : C -> Y -> Word X × C)
       (hypA : ∀ (x : X) -> Γ ⊢ₚ (h **) I (f x) === (gen x , I))
       (hypB : ∀ (c : C) {w w' : Word Y} -> w === w' ∈ Δ -> Γ ⊢ₚ (h **) c w === (h **) c w')
  where

    open Monoid-Equational

    -- Definition: A word is special if it doesn't leave the "I"-coset.
    special : Word Y -> Set
    special w = snd ((h **) I w) == I

    -- Lemma: ε is special.
    lemma-special-ε : special ε
    lemma-special-ε = Eq.refl

    -- Lemma: Special words are closed under multiplication.
    lemma-special-• : ∀ (w v : Word Y) -> special w -> special v -> special (w • v)
    lemma-special-• w v hyp-w hyp-v with (h **) I w | inspect ((h **) I) w
    lemma-special-• w v hyp-w hyp-v | (w' , c') | [ eq1 ] with (h **) c' v | inspect ((h **) c') v
    lemma-special-• w v hyp-w hyp-v | (w' , c') | [ eq1 ] | (v' , c'') | [ eq2 ] = lem3
      where
          lem1 : c' == I
          lem1 = hyp-w

          lem2 : (v' , c'') == (h **) I v
          lem2 =
            Eq.equational (v' , c'')
                    Eq.by Eq.symm eq2
                   equals (h **) c' v
                    Eq.by Eq.context (λ □ → (h **) □ v) lem1
                   equals (h **) I v

          lem3 : c'' == I
          lem3 = 
            Eq.equational c''
                    Eq.by Eq.context snd lem2
                   equals snd ((h **) I v)
                    Eq.by hyp-v
                   equals I

    -- Lemma: The image of f is special.
    lemma-special-f : ∀ (x : X) -> special (f x)
    lemma-special-f x = lemma-⊢ₚ-snd ((h **) I (f x)) (gen x , I) (hypA x)

    -- Lemma: The image of (f *) is special.
    lemma-special-f* : ∀ (w : Word X) -> special ((f *) w)
    lemma-special-f* (gen x) = lemma-special-f x
    lemma-special-f* ε = lemma-special-ε
    lemma-special-f* (w • v) = lemma-special-• ((f *) w) ((f *) v) (lemma-special-f* w) (lemma-special-f* v)

    -- Definition: For special words, there is a translation back.
    g : Word Y -> Word X
    g w = fst ((h **) I w)

    -- Lemma: g preserves ε.
    lemma-g-ε : g ε == ε
    lemma-g-ε = Eq.refl

    -- Lemma: g is a homomorphism on special words.
    lemma-g-• : ∀ (w v : Word Y) -> special w -> g (w • v) == g w • g v
    lemma-g-• w v hyp with (h **) I w | inspect ((h **) I) w
    lemma-g-• w v hyp | (w' , c') | [ eq1 ] with (h **) c' v | inspect ((h **) c') v
    lemma-g-• w v hyp | (w' , c') | [ eq1 ] | (v' , c'') | [ eq2 ] = lem6
      where
          lem1 : c' == I
          lem1 = hyp

          lem2 : (v' , c'') == (h **) I v
          lem2 =
            Eq.equational (v' , c'')
                    Eq.by eq2 Eq.reversed
                   equals (h **) c' v
                    Eq.by Eq.context (λ □ → (h **) □ v) lem1
                   equals (h **) I v

          lem4 : v' == g v
          lem4 = Eq.context fst lem2 

          lem6 : w' • v' == w' • g v
          lem6 = Eq.context (λ □ → w' • □) lem4

    -- Lemma: g is a left inverse of f *.
    lemma-a : ∀ (w : Word X) -> Γ ⊢ w === g ((f *) w)
    lemma-a (gen x) = lemma-⊢ₚ-fst (gen x , I) ((h **) I (f x)) (lemma-symm-⊢ₚ (hypA x))
    lemma-a ε = refl
    lemma-a (w • u) = 
      equational w • u
              by left lemma-a w
          equals g ((f *) w) • u
              by right lemma-a u
          equals g ((f *) w) • g ((f *) u)
              by refl' (lemma-g-• ((f *) w) ((f *) u) (lemma-special-f* w)) reversed
          equals g ((f *) w • (f *) u)

    -- Lemma: Hypothesis B can be extended from the elements of Δ to
    -- all consequences of Δ.
    lemma-hypB : ∀ (c : C) (u t : Word Y) -> Δ ⊢ u === t -> Γ ⊢ₚ (h **) c u === (h **) c t

    lemma-hypB c u t (axiom x) = hypB c x

    lemma-hypB c u .u refl = lemma-refl-⊢ₚ Eq.refl

    lemma-hypB c ((u • v) • w) (.u • (.v • .w)) assoc with (h **) c u
    ... | (u' , c') with (h **) c' v
    ... | (v' , c'') with (h **) c'' w
    ... | (w' , c''') = (assoc , Eq.refl)

    lemma-hypB c (ε • u) .u left-unit with (h **) c u
    ... | (u' , c') = (left-unit , Eq.refl)

    lemma-hypB c (u • ε) .u right-unit with (h **) c u
    ... | (u' , c') = (right-unit , Eq.refl)

    lemma-hypB c u t (symm hyp) = lemma-symm-⊢ₚ (lemma-hypB c t u hyp)

    lemma-hypB c u t (trans {v = v} hyp1 hyp2) = lemma-trans-⊢ₚ (lemma-hypB c u v hyp1) (lemma-hypB c v t hyp2)

    lemma-hypB c (u • u') (t • t') (cong hyp1 hyp2)
      with (h **) c u | (h **) c t | lemma-hypB c u t hyp1
    ... | (u'' , c') | (t'' , c'') | (ih1 , ih1')
      rewrite ih1'
      with (h **) c'' u' | (h **) c'' t' | lemma-hypB c'' u' t' hyp2
    ... | (u''' , c''') | (t''' , c'''') | (ih2 , ih2') = (cong ih1 ih2 , ih2')

    -- Lemma: g preserves relations.
    lemma-b : ∀ (u t : Word Y) -> Δ ⊢ u === t -> Γ ⊢ g u === g t
    lemma-b u t hyp with (h **) I u | (h **) I t | lemma-hypB I u t hyp
    ... | (u' , c') | (t' , c'') | (ih , ih') = ih

    -- Reidemeister-Schreier Theorem.
    reidemeister-schreier : (w v : Word X) -> Δ ⊢ (f *) w === (f *) v -> Γ ⊢ w === v
    reidemeister-schreier w v hyp =
      equational w
              by lemma-a w
          equals g ((f *) w)
              by lemma-b ((f *) w) ((f *) v) hyp
          equals g ((f *) v)
              by lemma-a v reversed
          equals v
