-- This file provides some combinators for equational reasoning
-- about words. It also provides some basic lemmas about monoids and
-- groups.

open import Proposition
open import Equality as Eq using (_==_)
open import Nat

open import Word

module Word-Lemmas where

-- ----------------------------------------------------------------------
-- * Combinators for equational reasoning

module Monoid-Equational where

  -- The following definitions provide a convenient syntax for writing
  -- equational proofs in monoids. To prove Γ ⊢ a0 === an, we can write:
  -- 
  -- equational a0
  --         by reason1
  --     equals a1
  --         by reason2
  --     equals a2
  --        ...
  --         by reasonn
  --     equals an
  --
  -- Note that the syntax is the same as that provided in Equality.agda,
  -- so depending on which type of equation one needs to reason about,
  -- only one or the other module should be opened.
  
  infix 2 equational_
  infixl 1 by-equals
  infixl 4 _reversed
  infixl 3 left_
  infixl 3 right_

  equational_ : ∀ {X : Set} {Γ : Context X} -> (w : Word X) -> Γ ⊢ w === w
  equational_ w = refl
  
  by-equals : ∀ {X : Set} {Γ : Context X} {w v : Word X} -> Γ ⊢ w === v -> (u : Word X) -> Γ ⊢ v === u -> Γ ⊢ w === u
  by-equals E u reason = trans E reason
  
  syntax by-equals E u reason = E by reason equals u
  
  _reversed : ∀ {X : Set} {Γ : Context X} {w v : Word X} -> Γ ⊢ w === v -> Γ ⊢ v === w
  _reversed = symm
  
  left_ : ∀ {X : Set} {Γ : Context X} {w v u : Word X} -> Γ ⊢ w === v -> Γ ⊢ w • u === v • u
  left_ E = cong E refl
  
  right_ : ∀ {X : Set} {Γ : Context X} {w v u : Word X} -> Γ ⊢ w === v -> Γ ⊢ u • w === u • v
  right_ E = cong refl E
  
  -- A version of reflexivity that can be used as an introduction rule.
  refl' : ∀ {X : Set} {Γ : Context X} {w v : Word X} -> w == v -> Γ ⊢ w === v
  refl' Eq.refl = refl
  
  -- A synonym for refl, so that we can say "by definition".
  definition : ∀ {X : Set} {Γ : Context X} {w} -> Γ ⊢ w === w
  definition = refl

-- ----------------------------------------------------------------------
-- * Basic lemmas about monoids

-- We take X and Γ to be fixed throughout the following lemmas.
module Monoid-Lemmas {X : Set} {Γ : Context X} where

  open Monoid-Equational

  -- The left unit is unique.
  lemma-left-unit-unique : {w : Word X} -> (∀ v -> Γ ⊢ w • v === v) -> Γ ⊢ w === ε
  lemma-left-unit-unique {w} p =
    equational w
            by right-unit reversed
        equals w • ε
            by p ε
        equals ε

  -- The right unit is unique.
  lemma-right-unit-unique : {w : Word X} -> (∀ v -> Γ ⊢ v • w === v) -> Γ ⊢ w === ε
  lemma-right-unit-unique {w} p =
    equational w
            by left-unit reversed
        equals ε • w
            by p ε
        equals ε

  -- Congruence for powers.
  lemma-cong-power : {w v : Word X} (n : Nat) -> Γ ⊢ w === v -> Γ ⊢ w ^ n === v ^ n
  lemma-cong-power zero hyp = refl
  lemma-cong-power (succ zero) hyp = hyp
  lemma-cong-power (succ (succ n)) hyp = cong hyp (lemma-cong-power (succ n) hyp)

  -- Recursive expansion of powers by repeated multiplication on the left.
  lemma-power-succ-left : {w : Word X} (n : Nat) -> Γ ⊢ w ^ (succ n) === w • w ^ n
  lemma-power-succ-left zero = right-unit reversed
  lemma-power-succ-left (succ n) = refl

  -- Recursive expansion of powers by repeated multiplication on the right.
  lemma-power-succ-right : {w : Word X} (n : Nat) -> Γ ⊢ w ^ (succ n) === w ^ n • w
  lemma-power-succ-right zero = left-unit reversed
  lemma-power-succ-right {w} (succ n) =
      equational w • w ^ succ n
              by right lemma-power-succ-right n
          equals w • (w ^ n • w)
              by assoc reversed
          equals (w • w ^ n) • w
              by left lemma-power-succ-left n reversed
          equals w ^ (succ n) • w

  -- A lemma about powers and multiplication.
  lemma-comm-power : ∀ {v w' w} -> (n : Nat) -> Γ ⊢ v • w === w' • v -> Γ ⊢ v • w ^ n === w' ^ n • v
  lemma-comm-power {v} {w'} {w} zero hyp =
      equational v • ε
              by right-unit
          equals v
              by left-unit reversed
          equals ε • v
  lemma-comm-power {v} {w'} {w} (succ n) hyp =
      equational v • w ^ succ n
              by right lemma-power-succ-left n
          equals v • (w • w ^ n)
              by assoc reversed
          equals (v • w) • w ^ n
              by left hyp
          equals (w' • v) • w ^ n
              by assoc
          equals w' • (v • w ^ n)
              by right lemma-comm-power n hyp
          equals w' • (w' ^ n • v)
              by assoc reversed
          equals (w' • w' ^ n) • v
              by left lemma-power-succ-left n reversed
          equals w' ^ succ n • v

  -- Commutativity of powers.
  lemma-comm-powers : ∀ {w v} -> (n m : Nat) -> Γ ⊢ w • v === v • w -> Γ ⊢ w ^ n • v ^ m === v ^ m • w ^ n
  lemma-comm-powers n m hyp = lemma-comm-power n (lemma-comm-power m hyp reversed) reversed

  -- Powers of products.
  lemma-product-power : ∀ {w v} -> (n : Nat) -> Γ ⊢ w • v === v • w -> Γ ⊢ (w • v) ^ n === w ^ n • v ^ n
  lemma-product-power zero hyp = right-unit reversed
  lemma-product-power {w} {v} (succ n) hyp =
      equational (w • v) ^ succ n
              by lemma-power-succ-left n
          equals (w • v) • (w • v) ^ n
              by right lemma-product-power n hyp
          equals (w • v) • (w ^ n • v ^ n)
              by assoc
          equals w • (v • (w ^ n • v ^ n))
              by right assoc reversed
          equals w • (v • w ^ n) • v ^ n
              by right left lemma-comm-power n (hyp reversed)
          equals w • (w ^ n • v) • v ^ n
              by right assoc
          equals w • (w ^ n • (v • v ^ n))
              by assoc reversed
          equals (w • w ^ n) • (v • v ^ n)
              by cong (lemma-power-succ-left n) (lemma-power-succ-left n) reversed
          equals w ^ succ n • v ^ succ n
  

  -- Power and change of basis.
  lemma-basis-change : ∀ {w w' v} -> (n : Nat) -> Γ ⊢ w' • w === ε -> Γ ⊢ (w • v • w') ^ succ n === w • v ^ succ n • w'
  lemma-basis-change {w} {w'} {v} 0 hyp = refl
  lemma-basis-change {w} {w'} {v} (succ n) hyp =
    equational (w • v • w') ^ succ (succ n)
            by definition
        equals (w • v • w') • (w • v • w') ^ succ n
            by right (lemma-basis-change n hyp)
        equals (w • v • w') • w • v ^ succ n • w'
            by left assoc reversed
        equals ((w • v) • w') • w • v ^ succ n • w'
            by assoc reversed
        equals (((w • v) • w') • w) • v ^ succ n • w'
            by left assoc
        equals ((w • v) • (w' • w)) • v ^ succ n • w'
            by left right hyp
        equals ((w • v) • ε) • v ^ succ n • w'
            by left right-unit
        equals (w • v) • v ^ succ n • w'
            by assoc
        equals w • v • v ^ succ n • w'
            by right assoc reversed
        equals w • (v • v ^ succ n) • w'
            by definition
        equals w • v ^ succ (succ n) • w'

-- ----------------------------------------------------------------------
-- ** Basic lemmas about groups

-- A context is group-like if it provides a left inverse to every
-- generator.
Grouplike : {X : Set} -> (Γ : Context X) -> Set
Grouplike {X} Γ = ∀ (x : X) -> ∃ λ (x' : Word X) -> Γ ⊢ x' • gen x === ε

-- We take X and Γ to be fixed throughout the following lemmas, and we
-- assume Γ is group-like.
module Group-Lemmas
       (X : Set)
       (Γ : Context X)
       (group-like : Grouplike Γ)
  where

  open Monoid-Equational

  infix 8 _⁻¹

  -- The inverse of a word.
  _⁻¹ : Word X -> Word X
  gen x ⁻¹ = fst (group-like x)
  ε ⁻¹ = ε
  (u • v) ⁻¹ = v ⁻¹ • u ⁻¹

  -- g ⁻¹ is a left inverse.
  lemma-left-inverse : {g : Word X} -> Γ ⊢ g ⁻¹ • g === ε
  lemma-left-inverse {gen x} = snd (group-like x)
  lemma-left-inverse {ε} = left-unit
  lemma-left-inverse {u • v} =
         equational (v ⁻¹ • u ⁻¹) • (u • v)
                 by assoc
             equals v ⁻¹ • (u ⁻¹ • (u • v))
                 by right assoc reversed
             equals v ⁻¹ • ((u ⁻¹ • u) • v)
                 by right left lemma-left-inverse
             equals v ⁻¹ • ε • v
                 by right left-unit
             equals v ⁻¹ • v
                 by lemma-left-inverse
             equals ε

  -- g ⁻¹ is a right inverse.
  lemma-right-inverse : {g : Word X} -> Γ ⊢ g • g ⁻¹ === ε
  lemma-right-inverse {g} = 
    equational g • (g ⁻¹)
            by left-unit reversed
        equals ε • (g • (g ⁻¹))
            by left lemma-left-inverse reversed
        equals ((g ⁻¹) ⁻¹ • g ⁻¹) • (g • (g ⁻¹))
            by assoc
        equals (g ⁻¹) ⁻¹ • (g ⁻¹ • (g • (g ⁻¹)))
            by right assoc reversed
        equals (g ⁻¹) ⁻¹ • ((g ⁻¹ • g) • g ⁻¹)
            by right left lemma-left-inverse
        equals (g ⁻¹) ⁻¹ • (ε • g ⁻¹)
            by right left-unit
        equals (g ⁻¹) ⁻¹ • g ⁻¹
            by lemma-left-inverse
        equals ε
  
  -- Left cancellation.
  lemma-left-cancel : {g h h' : Word X} -> Γ ⊢ g • h === g • h' -> Γ ⊢ h === h'
  lemma-left-cancel {g} {h} {h'} p = 
    equational h
               by left-unit reversed
        equals ε • h
               by left lemma-left-inverse reversed
        equals (g ⁻¹ • g) • h
               by assoc
        equals g ⁻¹ • (g • h)
               by right p
        equals g ⁻¹ • (g • h')
               by assoc reversed
        equals (g ⁻¹ • g) • h'
               by left lemma-left-inverse
        equals ε • h'
               by left-unit
        equals h'
  
  -- Right cancellation.
  lemma-right-cancel : {g g' h : Word X} -> Γ ⊢ g • h === g' • h -> Γ ⊢ g === g'
  lemma-right-cancel {g} {g'} {h} p = 
    equational g
               by right-unit reversed
        equals g • ε
               by right lemma-right-inverse reversed
        equals g • (h • h ⁻¹)
               by assoc reversed
        equals (g • h) • h ⁻¹
               by left p
        equals (g' • h) • h ⁻¹
               by assoc
        equals g' • (h • h ⁻¹)
               by right lemma-right-inverse
        equals g' • ε
               by right-unit
        equals g'
  
  -- Left inverses are unique.
  lemma-left-inverse-unique : {g h : Word X} -> Γ ⊢ h • g === ε -> Γ ⊢ h === g ⁻¹
  lemma-left-inverse-unique {g} {h} p = 
    equational h
               by right-unit reversed
        equals h • ε
               by right lemma-right-inverse reversed
        equals h • (g • g ⁻¹)
               by assoc reversed
        equals (h • g) • g ⁻¹
               by left p
        equals ε • g ⁻¹
               by left-unit
        equals g ⁻¹
  
  -- Right inverses are unique.
  lemma-right-inverse-unique : {g h : Word X} -> Γ ⊢ g • h === ε -> Γ ⊢ h === g ⁻¹
  lemma-right-inverse-unique {g} {h} p = 
    equational h
               by left-unit reversed
        equals ε • h
               by left lemma-left-inverse reversed
        equals (g ⁻¹ • g) • h
               by assoc
        equals g ⁻¹ • (g • h)
               by right p
        equals g ⁻¹ • ε
               by right-unit
        equals g ⁻¹

  -- Congruence for inverses.
  lemma-cong-inv : {g h : Word X} -> Γ ⊢ g === h -> Γ ⊢ g ⁻¹ === h ⁻¹
  lemma-cong-inv {g} {h} p = lemma-right-inverse-unique claim
     where
       claim : Γ ⊢ h • g ⁻¹ === ε
       claim = equational h • g ⁻¹
                       by left p reversed
                   equals g • g ⁻¹
                       by lemma-right-inverse
                   equals ε

  -- The inverse is involutive.
  lemma-inverse-involutive : {g : Word X} -> Γ ⊢ (g ⁻¹) ⁻¹ === g
  lemma-inverse-involutive {g} = lemma-right-inverse-unique lemma-left-inverse reversed
  
  -- The inverse of ε.
  lemma-unit-inverse : Γ ⊢ ε ⁻¹ === ε
  lemma-unit-inverse = refl
  
  -- The inverse of a product.
  lemma-product-inverse : ∀ {g h : Word X} -> Γ ⊢ (g • h) ⁻¹ === h ⁻¹ • g ⁻¹
  lemma-product-inverse = refl

  -- To show two words are equal, it is sufficient to show their
  -- inverses are equal.
  lemma-rule-inverse : ∀ {u v : Word X} -> Γ ⊢ u ⁻¹ === v ⁻¹ -> Γ ⊢ u === v
  lemma-rule-inverse {u} {v} hyp =
         equational u
                 by right-unit reversed
             equals u • ε
                 by right (lemma-left-inverse reversed)
             equals u • (v ⁻¹ • v)
                 by assoc reversed
             equals (u • v ⁻¹) • v
                 by left (right (hyp reversed))
             equals (u • u ⁻¹) • v
                 by left lemma-right-inverse
             equals ε • v
                 by left-unit
             equals v

  -- Commutativity of inverses. (Note: commutativity is the special
  -- case where v = v' and w = w')
  lemma-comm-inv : ∀ {v v' w w'} -> Γ ⊢ v • w === w' • v' -> Γ ⊢ v' • w ⁻¹ === w' ⁻¹ • v
  lemma-comm-inv {v} {v'} {w} {w'} hyp =
      equational v' • w ⁻¹
              by left-unit reversed
          equals ε • (v' • w ⁻¹)
              by left lemma-left-inverse reversed
          equals (w' ⁻¹ • w') • (v' • w ⁻¹)
              by assoc
          equals w' ⁻¹ • (w' • (v' • w ⁻¹))
              by right assoc reversed
          equals w' ⁻¹ • (w' • v') • w ⁻¹
              by right left hyp reversed
          equals w' ⁻¹ • (v • w) • w ⁻¹
              by right assoc
          equals w' ⁻¹ • v • (w • w ⁻¹)
              by assoc reversed
          equals (w' ⁻¹ • v) • (w • w ⁻¹)
              by right lemma-right-inverse
          equals (w' ⁻¹ • v) • ε
              by right-unit
          equals w' ⁻¹ • v

  -- Any equation can be reduced to a one-sided equation.
  lemma-one-sided : ∀ {w u} -> Γ ⊢ w • u ⁻¹ === ε -> Γ ⊢ w === u
  lemma-one-sided {w} {u} hyp =
      equational w
              by right-unit reversed
          equals w • ε
              by right lemma-left-inverse reversed
          equals w • (u ⁻¹ • u)
              by assoc reversed
          equals (w • u ⁻¹) • u
              by left hyp
          equals ε • u
              by left-unit
          equals u

-- ----------------------------------------------------------------------
-- * Subtheories

module Monoid-Subtheories where

  -- Γ is a subtheory of Γ' if each theorem of Γ is a theorem of Γ'.
  _⊆_ : {X : Set} -> (Γ Γ' : Context X) -> Set
  Γ ⊆ Γ' = ∀ {E} -> Γ ⊢ E -> Γ' ⊢ E
  
  -- Lemma: To show Γ ⊆ Γ', it suffices to show that each axiom of Γ
  -- is a theorem of Γ'.
  lemma-⊆ : ∀ {X : Set} {Γ Γ'} -> (∀ {E : Relation X} -> E ∈ Γ -> Γ' ⊢ E) -> Γ ⊆ Γ'
  lemma-⊆ hyp (axiom x) = hyp x
  lemma-⊆ hyp refl = refl
  lemma-⊆ hyp (symm p) = symm (lemma-⊆ hyp p)
  lemma-⊆ hyp (trans p q) = trans (lemma-⊆ hyp p) (lemma-⊆ hyp q)
  lemma-⊆ hyp (cong p q) = cong (lemma-⊆ hyp p) (lemma-⊆ hyp q)
  lemma-⊆ hyp assoc = assoc
  lemma-⊆ hyp left-unit = left-unit
  lemma-⊆ hyp right-unit = right-unit

