-- This module provides some convenient tactics for proving properties
-- of words, such as proving that two words are equal up to
-- associativity or commutativity, properties of inverses, support
-- for proofs involving rewrite systems, and others.

open import Boolean
open import Proposition
open import Equality as Eq using (_==_)
open import Inspect
open import Nat
open import Maybe
open import List hiding (split; lemma-split)

open import Word
open import Word-Lemmas

module Word-Tactics where

-- ----------------------------------------------------------------------
-- * Tactics for associativity

module Associative where

  -- This module provides some tactics for proving equations
  -- between words up to associativity and unit laws.
  --
  -- The tactic general-assoc can be used to automatically prove
  -- equalities of ground words (i.e., words that do not contain
  -- variables).  It can be used, for example, like this:
  --
  -- property : Γ ⊢ (H0 • ((CZ • S0) • ε)) === (H0 • (CZ • ε)) • S0
  -- property = general-assoc auto
  --
  -- The tactic special-assoc can be used to automatically prove
  -- equalities of non-ground words (i.e., those potentially
  -- containing variables). It can be used, for example, like this:
  --
  -- property : ∀ a b c -> Γ ⊢ (a • b) • (H0 • c) === a • (b • H0) • c
  -- property = special-assoc ((□ • □) • (□ • □)) (□ • (□ • □) • □) auto

  open Monoid-Equational

  -- Convert a list of generators to a word.
  word-of-list : ∀ {X} -> List X -> Word X
  word-of-list nil = ε
  word-of-list (x :: xs) = gen x • word-of-list xs

  -- Convert a word to a list of generators. Note that two words w, v
  -- are equal up to associativity and unit laws if and only if
  -- list-of-word w == list-of-word v.
  list-of-word : ∀ {X : Set} (w : Word X) -> List X
  list-of-word (gen x) = x :: nil
  list-of-word ε = nil
  list-of-word (w • u) = list-of-word w ++ list-of-word u

  -- Lemma: word-of-list of a homomorphism.
  lemma-append : ∀ {X Γ} -> (xs ys : List X) -> Γ ⊢ word-of-list xs • word-of-list ys === word-of-list (xs ++ ys)
  lemma-append nil ys = left-unit
  lemma-append (x :: xs) ys =
      equational (gen x • word-of-list xs) • word-of-list ys
              by assoc
          equals gen x • (word-of-list xs • word-of-list ys)
              by right lemma-append xs ys
          equals gen x • word-of-list (xs ++ ys)

  -- Lemma: word-of-list is the inverse of list-of-word, up to
  -- associativity.
  lemma-list-of-word : ∀ {X Γ} (w : Word X) -> Γ ⊢ w === word-of-list (list-of-word w)
  lemma-list-of-word (gen x) = right-unit reversed
  lemma-list-of-word ε = refl
  lemma-list-of-word (w • u) =
     equational w • u
             by left lemma-list-of-word w
         equals word-of-list (list-of-word w) • u
             by right lemma-list-of-word u
         equals word-of-list (list-of-word w) • word-of-list (list-of-word u)
             by lemma-append (list-of-word w) (list-of-word u)
         equals word-of-list (list-of-word w ++ list-of-word u)

  -- Lemma: list-of-word is the inverse of word-of-list.
  lemma-word-of-list : ∀ {X} (w : List X) -> w == list-of-word (word-of-list w)
  lemma-word-of-list nil = Eq.refl
  lemma-word-of-list (x :: w) = Eq.context (λ □ → x :: □) (lemma-word-of-list w)

  -- A tactic for arbitrary applications of associativity and unit
  -- laws to ground words. This reduces the problem of proving Γ ⊢ w
  -- === v to the problem of proving list-of-word w == list-of-word v,
  -- which can be done automatically if it is by reflexivity.
  general-assoc : ∀ {X Γ} {w v : Word X} -> list-of-word w == list-of-word v -> Γ ⊢ w === v
  general-assoc {X} {Γ} {w} {v} hyp =
     equational w
             by lemma-list-of-word w
         equals word-of-list (list-of-word w)
             by refl' (Eq.context word-of-list hyp)
         equals word-of-list (list-of-word v)
             by lemma-list-of-word v reversed
         equals v

  -- Another tactic: Here we need to justify a law only "up to
  -- associativity". For example, if we are given Γ ⊢ (X • Y) • Z
  -- === W, and we need to prove Γ ⊢ X • (Y • Z) === ε • W, we can
  -- apply the up-to-assoc tactic to the assumption to get the
  -- conclusion.
  up-to-assoc : ∀ {X Γ} {w v w' v' : Word X} -> (list-of-word w , list-of-word v) == (list-of-word w' , list-of-word v') -> Γ ⊢ w === v -> Γ ⊢ w' === v'
  up-to-assoc {X} {Γ} {w} {v} {w'} {v'} eqs deriv =
      equational w'
              by general-assoc (Eq.context fst eqs) reversed
          equals w
              by deriv
          equals v
              by general-assoc (Eq.context snd eqs)
          equals v'

  -- A convenient symbol to use in patterns such as (□ • □) • □.
  □ : Word True
  □ = gen <>

  -- A version of list-of-word that uses another "pattern" word to
  -- guide the conversion. Any subwords corresponding to occurrences
  -- of □ in the pattern word are not reduced further.
  list-of-word-special : ∀ {X} -> Word X -> Word True -> List (Word X)
  list-of-word-special w (gen <>) = w :: nil
  list-of-word-special (gen x) ε = gen x :: nil
  list-of-word-special (gen x) (p • q) = gen x :: nil
  list-of-word-special ε ε = nil
  list-of-word-special ε (p • q) = nil
  list-of-word-special (w • v) ε = list-of-word-special w ε ++ list-of-word-special v ε
  list-of-word-special (w • v) (p • q) = list-of-word-special w p ++ list-of-word-special v q

  -- Definition: Flatten a word of words into a word.
  flatten-word : ∀ {X} -> Word (Word X) -> Word X
  flatten-word (gen w) = w
  flatten-word ε = ε
  flatten-word (w • v) = flatten-word w • flatten-word v

  -- Lemma: If two words are equivalent, then so are their flattenings.
  lemma-flatten-word : ∀ {X Γ} -> {xs ys : Word (Word X)} -> ∅ ⊢ xs === ys -> Γ ⊢ flatten-word xs === flatten-word ys 
  lemma-flatten-word (axiom ())
  lemma-flatten-word refl = refl
  lemma-flatten-word (symm hyp) = symm (lemma-flatten-word hyp)
  lemma-flatten-word (trans hyp hyp₁) = trans (lemma-flatten-word hyp) (lemma-flatten-word hyp₁)
  lemma-flatten-word (cong hyp hyp₁) = cong (lemma-flatten-word hyp) (lemma-flatten-word hyp₁)
  lemma-flatten-word assoc = assoc
  lemma-flatten-word left-unit = left-unit
  lemma-flatten-word right-unit = right-unit

  -- Lemma: word-of-list is the inverse of list-of-word-special, up to
  -- flattening.
  lemma-list-of-word-special : ∀ {X Γ} (w : Word X) (p : Word True) -> Γ ⊢ flatten-word (word-of-list (list-of-word-special w p)) === w
  lemma-list-of-word-special w (gen <>) = right-unit
  lemma-list-of-word-special (gen x) ε = right-unit
  lemma-list-of-word-special (gen x) (p • q) = right-unit
  lemma-list-of-word-special ε ε = refl
  lemma-list-of-word-special ε (p • q) = refl
  lemma-list-of-word-special (w • v) ε =
      equational flatten-word (word-of-list (list-of-word-special w ε ++ list-of-word-special v ε))
              by lemma-flatten-word (lemma-append (list-of-word-special w ε) (list-of-word-special v ε)) reversed
          equals flatten-word (word-of-list (list-of-word-special w ε) • word-of-list (list-of-word-special v ε))
              by refl
          equals flatten-word (word-of-list (list-of-word-special w ε)) • flatten-word (word-of-list (list-of-word-special v ε))
              by cong (lemma-list-of-word-special w ε) (lemma-list-of-word-special v ε)
          equals w • v

  lemma-list-of-word-special (w • v) (p • q) = 
      equational flatten-word (word-of-list (list-of-word-special w p ++ list-of-word-special v q))
              by lemma-flatten-word (lemma-append (list-of-word-special w p) (list-of-word-special v q)) reversed
          equals flatten-word (word-of-list (list-of-word-special w p) • word-of-list (list-of-word-special v q))
              by refl
          equals flatten-word (word-of-list (list-of-word-special w p)) • flatten-word (word-of-list (list-of-word-special v q))
              by cong (lemma-list-of-word-special w p) (lemma-list-of-word-special v q)
          equals w • v

  -- A "guided" version of general associativity. Unlike
  -- general-assoc, this also works for non-ground words, i.e., words
  -- that contain variables. This is achieved by giving a pair of
  -- "pattern" words containing □ in all places that should not be
  -- reduced.
  --
  -- For example, to prove Γ ⊢ (a • b) • (c • d) === a • (b • c) • d,
  -- use
  -- special-assoc ((□ • □) • (□ • □)) (□ • (□ • □) • □) auto.
  --
  -- This even works when a, b, c, d are variables. For added
  -- convenience, ε can be used as a pattern for any determinate
  -- sub-word (i.e., a subword that contains no variables).
  special-assoc : ∀ {X Γ} {w v : Word X} (p q : Word True) -> list-of-word-special w p == list-of-word-special v q -> Γ ⊢ w === v
  special-assoc {w = w} {v = v} p q hyp =
      equational w
              by lemma-list-of-word-special w p reversed
          equals flatten-word (word-of-list (list-of-word-special w p))
              by refl' (Eq.context (λ □ → flatten-word (word-of-list □)) hyp)
          equals flatten-word (word-of-list (list-of-word-special v q))
              by lemma-list-of-word-special v q
          equals v

-- ----------------------------------------------------------------------
-- * Tactics for commutativity

-- Definition: two generators commute under Γ.
commutes : ∀ {X} -> Context X -> X -> X -> Set
commutes Γ x y = Γ ⊢ gen x • gen y === gen y • gen x

-- Definition: two words commute under Γ.
commutes-word : ∀ {X} -> Context X -> (u v : Word X) -> Set
commutes-word Γ u v = Γ ⊢ u • v === v • u

module Commuting
       (X : Set)
       (Γ : Context X)
       (comm : (x y : X) -> Maybe (commutes Γ x y))
       (less : X -> X -> Boolean)
  where

  -- This module provides some tactics for proving equations between
  -- words up to commutativity. We assume given a set X of generators,
  -- a set Γ of relations, and a semi-decidable commutativity relation
  -- (i.e., given generators x and y, comm x y either returns a proof
  -- that x and y commute, or Nothing). We also assume a total
  -- ordering 'less' on generators, which is used to find canonical
  -- representatives of words up to commutativity.
  --
  -- The tactic general-comm can be used to automatically prove
  -- equalities of ground words up to commutativity. It can be used,
  -- for example, as follows. Here, we assume that Z commutes with X
  -- and Y, but X does not commute with Y:
  --
  -- property : Γ ⊢ X • Y • Z === Z • X • Y
  -- property = general-comm auto
  --
  -- Note that this tactic also includes associativity.

  open Monoid-Equational
  open Associative

  open List.All

  -- ----------------------------------------------------------------------
  -- ** Canonical forms

  -- The primary engine of the commutativity tactic is the function
  -- comm-canonical, which computes the canonical form of a word
  -- modulo commutativity. The canonical form is defined to be the
  -- "smallest" word that is equivalent to the given one. Here,
  -- "smallest" is taken with respect to the reverse (or Hebrew)
  -- lexicographic order.
  --
  -- For example, consider generators A, B, C, D such that A commutes
  -- with C and D; B commutes with D; and no other pair of generators
  -- commutes. Assume A < B < C < D. Then the canonical form of
  -- B,A,D,C,A,B is D,B,C,A,A,B.
  --
  -- The reason we use the reverse lexicographic order is that it
  -- permits a convenient and efficient recursive algorithm for
  -- computing canonical forms. Namely, we can compute the canonical
  -- form of a non-empty list of generators x::xs by first computing
  -- the canonical form of xs, then inserting x in the correct
  -- position.

  -- Auxiliary function: Given ys, x, and zs, return
  -- (reverse zs1 @ ys , zs2), where zs = zs1 @ zs2 and zs1 is the
  -- longest prefix of zs all of whose elements commute with x.
  split : List X -> X -> List X -> List X × List X
  split ys x nil = (ys , nil)
  split ys x (z :: zs) with comm x z
  split ys x (z :: zs) | Nothing = (ys , z :: zs)
  split ys x (z :: zs) | Just _ = split (z :: ys) x zs

  -- Auxiliary function: Given ys, x, and zs, return reverse ys' @ zs,
  -- where ys' is the lexicographically smallest word that can be
  -- obtained by inserting x in ys. We assume that reverse ys @ zs is
  -- a canonical form.
  unsplit : List X -> X -> List X -> List X
  unsplit nil x zs = x :: zs
  unsplit (y :: ys) x zs with less x y
  unsplit (y :: ys) x zs | true = unsplit ys x (y :: zs)
  unsplit (y :: ys) x zs | false = reverse-append (y :: ys) (x :: zs)

  -- Given x and xs, where xs is already a canonical form, return the
  -- canonical form of x :: xs.
  insert : X -> List X -> List X
  insert x xs with split nil x xs
  insert x xs | (ys , zs) = unsplit ys x zs

  -- Return the canonical form of a word, modulo commutativity of
  -- generators.  Of all equivalent words, the canonical form is the
  -- least one in the reverse lexicographic order.
  comm-canonical : List X -> List X
  comm-canonical nil = nil
  comm-canonical (x :: xs) = insert x (comm-canonical xs)

  -- ----------------------------------------------------------------------
  -- ** Properties of canonical forms.

  -- We prove various technical properties of the functions split,
  -- reverse-append, unsplit, and insert. These will be needed for
  -- proving that every word is equivalent to its canonical form.

  lemma-split1 : (ys : List X) -> (x : X) -> (zs : List X) -> ∀ {ys' zs'} -> split ys x zs == (ys' , zs') -> Γ ⊢ word-of-list (reverse-append ys zs) === word-of-list (reverse-append ys' zs')
  lemma-split1 ys x nil Eq.refl = refl
  lemma-split1 ys x (z :: zs) eq with comm x z
  lemma-split1 ys x (z :: zs) Eq.refl | Nothing = refl
  lemma-split1 ys x (z :: zs) eq | Just wit = lemma-split1 (z :: ys) x zs eq

  lemma-split2 : (ys : List X) -> (x : X) -> (zs : List X) -> All[ y ∈ ys ] commutes Γ x y -> ∀ {ys' zs'} -> split ys x zs == (ys' , zs') -> All[ y ∈ ys' ] commutes Γ x y
  lemma-split2 ys x nil ih Eq.refl = ih -- ih
  lemma-split2 ys x (z :: zs) ih eq with comm x z
  lemma-split2 ys x (z :: zs) ih Eq.refl | Nothing = ih -- ih
  lemma-split2 ys x (z :: zs) ih eq | Just wit = lemma-split2 (z :: ys) x zs (wit :: ih) eq

  lemma-reverse-append-cong : (ys : List X) -> (zs zs' : List X) -> Γ ⊢ word-of-list zs === word-of-list zs' -> Γ ⊢ word-of-list (reverse-append ys zs) === word-of-list (reverse-append ys zs')
  lemma-reverse-append-cong nil zs zs' hyp = hyp
  lemma-reverse-append-cong (y :: ys) zs zs' hyp = lemma-reverse-append-cong ys (y :: zs) (y :: zs') (right hyp)

  lemma-reverse-append : (ys : List X) -> (x : X) -> (zs : List X) -> All[ y ∈ ys ] commutes Γ x y -> Γ ⊢ word-of-list (x :: reverse-append ys zs) === word-of-list (reverse-append ys (x :: zs))
  lemma-reverse-append nil x zs [] = refl
  lemma-reverse-append (y :: ys) x zs (h :: hyp) =
    let ih : Γ ⊢ word-of-list (x :: reverse-append ys (y :: zs)) === word-of-list (reverse-append ys (x :: y :: zs))
        ih = lemma-reverse-append ys x (y :: zs) hyp

        claim : Γ ⊢ gen x • gen y • word-of-list zs === gen y • gen x • word-of-list zs
        claim = equational gen x • gen y • word-of-list zs
                        by assoc reversed
                    equals (gen x • gen y) • word-of-list zs
                        by left h
                    equals (gen y • gen x) • word-of-list zs
                        by assoc
                    equals gen y • gen x • word-of-list zs
    in
    equational word-of-list (x :: reverse-append ys (y :: zs))
            by ih
        equals word-of-list (reverse-append ys (x :: y :: zs))
            by lemma-reverse-append-cong ys (x :: y :: zs) (y :: x :: zs) claim
        equals word-of-list (reverse-append ys (y :: x :: zs))
       
  lemma-unsplit : (ys : List X) -> (x : X) -> (zs : List X) -> All[ y ∈ ys ] commutes Γ x y -> Γ ⊢ word-of-list (x :: reverse-append ys zs) === word-of-list (unsplit ys x zs)
  lemma-unsplit nil x zs hyp = refl
  lemma-unsplit (y :: ys) x zs (h :: hyp) with less x y
  lemma-unsplit (y :: ys) x zs (h :: hyp) | true = lemma-unsplit ys x (y :: zs) hyp
  lemma-unsplit (y :: ys) x zs (h :: hyp) | false = lemma-reverse-append (y :: ys) x zs (h :: hyp)

  lemma-insert : (x : X) -> (xs : List X) -> Γ ⊢ word-of-list (x :: xs) === word-of-list (insert x xs)
  lemma-insert x xs with split nil x xs | inspect (split nil x) xs
  ... | (ys , zs) | [ eq ] =
      equational word-of-list (x :: xs)
              by right lemma-split1 nil x xs eq
          equals word-of-list (x :: reverse-append ys zs)
              by lemma-unsplit ys x zs (lemma-split2 nil x xs [] eq)
          equals word-of-list (unsplit ys x zs)

  -- Soundness of comm-canonical: every word is equivalent to its
  -- canonical form.
  lemma-comm-canonical : (xs : List X) -> Γ ⊢ word-of-list xs === word-of-list (comm-canonical xs)
  lemma-comm-canonical nil = refl
  lemma-comm-canonical (x :: xs) =
    equational word-of-list (x :: xs)
            by right (lemma-comm-canonical xs)
        equals word-of-list (x :: comm-canonical xs)
            by lemma-insert x (comm-canonical xs)
        equals word-of-list (insert x (comm-canonical xs))

  -- ----------------------------------------------------------------------
  -- ** The general-comm tactic

  -- A tactic for showing that two ground words are equivalent up to
  -- commutativity (and associativity). This works by converting both
  -- words to canonical form and checking that they are equal.
  general-comm : ∀ {w v : Word X} -> comm-canonical (list-of-word w) == comm-canonical (list-of-word v) -> Γ ⊢ w === v
  general-comm {w} {v} hyp =
   equational w
           by lemma-list-of-word w
       equals word-of-list (list-of-word w)
           by lemma-comm-canonical (list-of-word w)
       equals word-of-list (comm-canonical (list-of-word w))
           by refl' (Eq.context word-of-list hyp)
       equals word-of-list (comm-canonical (list-of-word v))
           by lemma-comm-canonical (list-of-word v) reversed
       equals word-of-list (list-of-word v)
           by lemma-list-of-word v reversed
       equals v

-- ----------------------------------------------------------------------
-- * Some tactics for inverse.

module Inverse
       (X : Set)
       (Γ : Context X)
       (group-like : Grouplike Γ)
  where

  -- This module provides a tactic inverse-eq, which can be used to
  -- prove that two words are equal by proving that their inverses are
  -- equal. It is similar to lemma-rule-inverse (from Word-Lemmas),
  -- except that the final ε on the inverses is omitted.

  open Monoid-Equational
  open Associative
  open Group-Lemmas X Γ group-like public

  -- An alternative version of word-of-list that omits the final ε.
  -- Sometimes this is preferable for cosmetic reasons.
  word-of-list-alt : ∀ {X} -> List X -> Word X
  word-of-list-alt nil = ε
  word-of-list-alt (x :: nil) = gen x
  word-of-list-alt (x :: y :: xs) = gen x • word-of-list-alt (y :: xs)

  -- Lemma: word-of-list-alt is equivalent to word-of-list.
  lemma-word-of-list-alt : ∀ {X Γ} -> (xs : List X) -> Γ ⊢ word-of-list xs === word-of-list-alt xs
  lemma-word-of-list-alt nil = refl
  lemma-word-of-list-alt (x :: nil) = right-unit
  lemma-word-of-list-alt (x :: y :: xs) = right lemma-word-of-list-alt (y :: xs)

  -- A version of lemma-rule-inverse that flattens the inverse, for
  -- convenience.
  inverse-eq : {u v : Word X} -> Γ ⊢ word-of-list-alt (list-of-word (u ⁻¹)) === word-of-list-alt (list-of-word (v ⁻¹)) -> Γ ⊢ u === v
  inverse-eq {u} {v} hyp = lemma-rule-inverse claim
    where
      claim : Γ ⊢ u ⁻¹ === v ⁻¹
      claim =
        equational u ⁻¹
                by lemma-list-of-word (u ⁻¹)
            equals word-of-list (list-of-word (u ⁻¹))
                by lemma-word-of-list-alt (list-of-word (u ⁻¹))
            equals word-of-list-alt (list-of-word (u ⁻¹))
                by hyp
            equals word-of-list-alt (list-of-word (v ⁻¹))
                by lemma-word-of-list-alt (list-of-word (v ⁻¹)) reversed
            equals word-of-list (list-of-word (v ⁻¹))
                by lemma-list-of-word (v ⁻¹) reversed
            equals v ⁻¹

-- ----------------------------------------------------------------------
-- * Some tactics for proofs by rewriting

module Rewriting
  where

  -- This module provides some tactics for proving equations between
  -- words by using a rewrite system. To use these tactics, a
  -- particular set of single-step rewrite rules must be supplied in
  -- the form of a Step-Function. The tactic general-rewrite in the
  -- submodule Step can be used to prove that two words are equal by
  -- rewriting them both to normal form. The tactic general-rewrite in
  -- the submodule Step-With-Standardization does the same, except it
  -- also applies a user-supplied standardization function before and
  -- after each rewrite step. This is typically used to implement
  -- rewriting up to commutativity.

  open Monoid-Equational
  open Associative

  infixl 4 _then_

  -- ----------------------------------------------------------------------
  -- ** Step functions

  -- The type for step functions. A step function inputs a word and
  -- outputs either Nothing (if no rewrite rule can be applied), or
  -- else a new word (the result of applying a rewrite rule), together
  -- with a proof that the original word and the rewritten word are
  -- equivalent.  For efficiency, words are here implemented as lists
  -- of generators.
  Step-Function : (X : Set) -> (Γ : Context X) -> Set
  Step-Function X Γ = (xs : List X) -> Maybe (∃ λ (xs' : List X) -> Γ ⊢ word-of-list xs === word-of-list xs')

  -- Sequentially combine two step functions. If the first step
  -- function successfully rewrites the word, use it; otherwise, use
  -- the second step function.
  _then_ : ∀ {X Γ} -> Step-Function X Γ -> Step-Function X Γ -> Step-Function X Γ
  (step1 then step2) xs with step1 xs
  ...                   | Just res = Just res
  ...                   | Nothing = step2 xs

  -- Close a step function under congruence. In other words, input a
  -- step function that only rewrites the head of a list, and return a
  -- step function that rewrites anywhere within the list.
  step-cong : ∀ {X Γ} -> Step-Function X Γ -> Step-Function X Γ
  step-cong step nil = Nothing
  step-cong step (h :: t) with step (h :: t)
  step-cong step (h :: t) | Just res = Just res
  step-cong step (h :: t) | Nothing with step-cong step t
  step-cong step (h :: t) | Nothing | Just (t' , hyp) = Just (h :: t' , (right hyp))
  step-cong step (h :: t) | Nothing | Nothing = Nothing

  -- ----------------------------------------------------------------------
  -- ** Multistep rewriting

  module Step
         {X : Set}
         {Γ : Context X}
         (step : Step-Function X Γ)
    where

    -- This module provides the tactic general-rewrite. Given a step
    -- function, this tactic can be used to prove that two words are
    -- equal by repeatedly applying the step function until the words
    -- are in normal form, then checking that they are equal.

    -- Users of this module typically specialize the module to a
    -- particular rewrite function. This can be done, for example, as
    -- follows:
    -- 
    -- module My-Rewrite = Rewriting.Step (step-cong my-stepfunction) renaming (general-rewrite to my-rewrite)
    --
    -- The tactic can then be used as follows:
    --
    -- open My-Rewrite
    -- property : Γ ⊢ A • B • C === D • E
    -- property = my-rewrite 100 auto

    open List.Strict

    -- Rewrite the given word until a normal form is reached, or up to
    -- a maximum of n rewrite steps. The parameter n is needed to
    -- ensure termination.
    multistep : (n : Nat) -> List X -> List X
    multistep zero xs = xs
    multistep (succ n) xs with strict xs step
    multistep (succ n) xs | Nothing = xs
    multistep (succ n) xs | Just (xs' , _) = multistep n xs'

    -- Lemma: multistep rewriting returns an equivalent word.
    lemma-multistep : (n : Nat) -> (xs : List X) -> Γ ⊢ word-of-list xs === word-of-list (multistep n xs)
    lemma-multistep zero xs = refl
    lemma-multistep (succ n) xs with strict xs step | inspect (strict xs) step
    lemma-multistep (succ n) xs | Nothing | _ = refl
    lemma-multistep (succ n) xs | Just (xs' , hyp) | [ eq ] =
      equational word-of-list xs
              by hyp
          equals word-of-list xs'
              by lemma-multistep n xs'
          equals word-of-list (multistep n xs')

    -- Return the entire rewrite sequence. This is useful for
    -- debugging rewriting strategies.
    multistep-trace : (n : Nat) -> List X -> List (List X)
    multistep-trace zero xs = xs :: nil
    multistep-trace (succ n) xs with step xs
    multistep-trace (succ n) xs | Nothing = xs :: nil
    multistep-trace (succ n) xs | Just (xs' , _) = xs :: multistep-trace n xs'

    -- A tactic for proving equality of ground words based on a
    -- rewrite relation. The parameter n limits the number of rewrite
    -- steps applied, and is needed to ensure termination. If n is too
    -- small, the tactic may fail.
    general-rewrite : (n : Nat) -> {w u : Word X} -> multistep n (list-of-word w) == multistep n (list-of-word u) -> Γ ⊢ w === u
    general-rewrite n {w} {u} eq =
      let ws = list-of-word w
          us = list-of-word u
      in 
        equational w
                by lemma-list-of-word w
            equals word-of-list ws
                by lemma-multistep n ws
            equals word-of-list (multistep n ws)
                by refl' (Eq.context word-of-list eq)
            equals word-of-list (multistep n us)
                by lemma-multistep n us reversed
            equals word-of-list us
                by lemma-list-of-word u reversed
            equals u

  -- ----------------------------------------------------------------------
  -- ** Multistep rewriting with standardization
  module Step-With-Standardization
         {X : Set}
         {Γ : Context X}
         (step : Step-Function X Γ)
         (standardize : List X -> List X)
         (lemma-standardize : (xs : List X) ->  Γ ⊢ word-of-list xs === word-of-list (standardize xs))
    where

    -- This module provides another version of the tactic
    -- general-rewrite that also applies a standardization function
    -- after each step. This is typically used to implement rewriting
    -- up to commutativity.
    --
    -- Users of this module typically specialize the module to a
    -- particular rewrite function and standardization function. This
    -- can be done, for example, as follows:
    --
    -- module My-Rewrite = Rewriting.Step (step-cong my-stepfunction) my-standardize my-lemma-standardize renaming (general-rewrite to my-rewrite)
    --
    -- The tactic can then be used as follows:
    --
    -- open My-Rewrite
    -- property : Γ ⊢ A • B • C === D • E
    -- property = my-rewrite 100 auto

    open List.Strict

    mutual
      -- Rewrite the given word until a normal form is reached, or up
      -- to a maximum of n rewrite steps. The parameter n is needed to
      -- ensure termination. The given standardization function is
      -- applied before and after every rewrite step.
      multistep : (n : Nat) -> List X -> List X
      multistep n xs = multistep-st n (standardize xs)

      -- Auxiliary function: multistep-st assumes that the input is
      -- already standardized.
      multistep-st : (n : Nat) -> List X -> List X
      multistep-st zero xs = xs
      multistep-st (succ n) xs with strict xs step
      multistep-st (succ n) xs | Nothing = xs
      multistep-st (succ n) xs | Just (xs' , _) = multistep n xs'

    mutual
      -- Lemma: multistep rewriting returns an equivalent word.
      lemma-multistep : (n : Nat) -> (xs : List X) -> Γ ⊢ word-of-list xs === word-of-list (multistep n xs)
      lemma-multistep n xs =
        let xs' = standardize xs
        in
          equational word-of-list xs
                  by lemma-standardize xs
              equals word-of-list xs'
                  by lemma-multistep-st n xs'
              equals word-of-list (multistep-st n xs')
                  by refl
              equals word-of-list (multistep n xs)

      lemma-multistep-st : (n : Nat) -> (xs : List X) -> Γ ⊢ word-of-list xs === word-of-list (multistep-st n xs)
      lemma-multistep-st zero xs = refl
      lemma-multistep-st (succ n) xs with strict xs step | inspect (strict xs) step
      lemma-multistep-st (succ n) xs | Nothing | _ = refl
      lemma-multistep-st (succ n) xs | Just (xs' , hyp) | [ eq ] =
        equational word-of-list xs
                by hyp
            equals word-of-list xs'
                by lemma-multistep n xs'
            equals word-of-list (multistep n xs')

    -- Return the entire rewrite sequence. Useful for debugging
    -- rewriting strategies. Note: in the output, standardization
    -- steps will alternate with rewrite steps, even if the
    -- standardization is the identity.
    mutual
      multistep-trace : (n : Nat) -> List X -> List (List X)
      multistep-trace n xs = xs :: multistep-trace-st n (standardize xs)

      multistep-trace-st : (n : Nat) -> List X -> List (List X)
      multistep-trace-st zero xs = xs :: nil
      multistep-trace-st (succ n) xs with step xs
      multistep-trace-st (succ n) xs | Nothing = xs :: nil
      multistep-trace-st (succ n) xs | Just (xs' , _) = xs :: multistep-trace n xs'
  
    -- A tactic for proving equality of ground words based on a
    -- rewrite relation and a standardization function. The parameter
    -- n limits the number of rewrite steps applied, and is needed to
    -- ensure termination. If n is too small, the tactic may fail.
    general-rewrite : (n : Nat) -> {w u : Word X} -> multistep n (list-of-word w) == multistep n (list-of-word u) -> Γ ⊢ w === u
    general-rewrite n {w} {u} eq =
      let ws = list-of-word w
          us = list-of-word u
      in 
        equational w
                by lemma-list-of-word w
            equals word-of-list ws
                by lemma-multistep n ws
            equals word-of-list (multistep n ws)
                by refl' (Eq.context word-of-list eq)
            equals word-of-list (multistep n us)
                by lemma-multistep n us reversed
            equals word-of-list us
                by lemma-list-of-word u reversed
            equals u

-- ----------------------------------------------------------------------
-- * Some tactics for applying axioms inside words

module InContext where

  -- This module provides some convenient tactics for applying axiom x
  -- at position n in some word. For the purpose of the following
  -- summary, assume there is an axiom
  --
  -- axiom : Γ ⊢ A • B === C • D • E.
  --
  -- The tactic at-head can be used to apply an axiom at the beginning
  -- of a list-like word. A word is list-like if it is of the form A •
  -- (B • (C • ε)), where A, B, C are generators. For example, this
  -- tactic can be used as follows:
  --
  -- property1 : Γ ⊢ A • B • X • Y • ε === C • D • E • X • Y • ε.
  -- property1 = at-head axiom
  --
  -- The tactic "in-context" can be used much more generally to apply
  -- an axiom in some specified position in a word. For example,
  --
  -- property2 : Γ ⊢ X • Y • Z • A • B • W === X • Y • Z • C • D • E • W
  -- property2 = in-context 3 2 auto axiom
  --
  -- Unlike at-head, the components X, Y, Z, A, B, etc., do not need to
  -- be generators; they can be words. The first two arguments of
  -- in-context specify the number of components of the left-hand side
  -- to be skipped (here, X, Y, Z), and the number of components of
  -- the left-hand-side to apply the axiom to (here, A, B).
  --
  -- The tactic in-context is used extensively in computer-generated
  -- proofs, such as the ones found in Equation1.agda, ...,
  -- Equation46.agda.
  --
  -- We also provide a tactic "rewrite-in-context" that is similar to
  -- in-context, but it applies rewriting to the selected subword,
  -- rather than an axiom. Specifically, given a multistep rewrite
  -- function 'multistep' and its associated lemma 'lemma-multistep',
  -- the tactic is used like this:
  --
  -- property3 : Γ ⊢ X • Y • Z • A • B • W === X • Y • Z • C • D • E • W
  -- property3 = rewrite-in-context multistep lemma-multistep 3 2 100 auto
  --
  -- Here, the arguments 3 and 2 are as above, i.e., the pick out the
  -- subword at position 3 and length 2 from the left-hand side (in
  -- this case, A • B). The argument 100 specifies a bound on the
  -- number of rewrite steps to perform. The tactic will then attempt
  -- to rewrite the words A • B and C • D • E to a common normal form.
  -- The tactic fails if this is not possible in the given number of
  -- steps, or if the rest of the words do not agree.


  open Monoid-Equational
  open Associative

  -- Translate the word u₁ • u₂ • ... • uₙ to the list [u₁, u₂, ..., uₙ].
  -- This differs from 'list-of-word' in that u₁, ..., uₙ are words, 
  -- not generators, i.e., we do not flatten subwords.
  list-of-word2 : ∀ {X} -> Word X -> List (Word X)
  list-of-word2 (gen x) = gen x :: nil
  list-of-word2 ε = ε :: nil
  list-of-word2 (u • w) = u :: list-of-word2 w
    
  -- Translate the list [u₁, u₂, ..., uₙ] to the word u₁ • u₂ • ... • uₙ.
  word-of-list2 : ∀ {X} -> List (Word X) -> Word X
  word-of-list2 nil = ε
  word-of-list2 (u :: nil) = u
  word-of-list2 (u :: v :: us) = u • word-of-list2 (v :: us)
  
  -- Lemma: word-of-list2 is an inverse of list-of-word2.
  lemma-list-of-word2 : ∀ {X} -> (u : Word X) -> word-of-list2 (list-of-word2 u) == u
  lemma-list-of-word2 (gen x) = Eq.refl
  lemma-list-of-word2 ε = Eq.refl
  lemma-list-of-word2 (u • gen x) = Eq.refl
  lemma-list-of-word2 (u • ε) = Eq.refl
  lemma-list-of-word2 (u • (w • v)) = Eq.context (λ □ → u • □) (lemma-list-of-word2 (w • v))
  
  -- A wrapper around List.split3: split a list of words into 3 parts,
  -- where the first has length up to n and the second has length up
  -- to m.
  mysplit : ∀ {X} -> Nat -> Nat -> List (Word X) -> List X × Word X × List X
  mysplit n m us with split3 n m us
  ... | (xs , ys , zs) = (list-of-word (word-of-list2 xs) , word-of-list2 ys , list-of-word (word-of-list2 zs))
  
  -- Lemma: word-of-list2 respects concatenation of lists, up to
  -- associativity.
  lemma-append2 : ∀ {X Γ} (us ws : List (Word X)) -> Γ ⊢ word-of-list2 us • word-of-list2 ws === word-of-list2 (us ++ ws)
  lemma-append2 nil ws = left-unit
  lemma-append2 (u :: nil) nil = right-unit
  lemma-append2 (u :: nil) (w :: ws) = refl
  lemma-append2 (u :: u' :: us) ws =
    equational (u • word-of-list2 (u' :: us)) • word-of-list2 ws
            by assoc
        equals u • (word-of-list2 (u' :: us) • word-of-list2 ws)
            by right lemma-append2 (u' :: us) ws
        equals u • word-of-list2 (u' :: (us ++ ws))

  -- Lemma: Concatenating the output of mysplit is equivalent to the
  -- original word, up to associativity.
  lemma-mysplit : ∀ {X} {Γ : Context X} n m us {x y z} -> mysplit n m us == (x , y , z) -> Γ ⊢ word-of-list x • y • word-of-list z === word-of-list2 us
  lemma-mysplit n m us {x} {y} {z} hyp with split3 n m us | inspect (split3 n m) us
  ... | (xs , ys , zs) | [ eq ] =
    let hyp1 = Eq.context fst hyp
        hyp2 = Eq.context fst (Eq.context snd hyp)
        hyp3 = Eq.context snd (Eq.context snd hyp)
    in
    equational word-of-list x • y • word-of-list z
            by left refl' (Eq.context word-of-list hyp1) reversed
        equals word-of-list (list-of-word (word-of-list2 xs)) • y • word-of-list z
            by left lemma-list-of-word (word-of-list2 xs) reversed
        equals word-of-list2 xs • y • word-of-list z
            by right right refl' (Eq.context word-of-list hyp3) reversed
        equals word-of-list2 xs • y • word-of-list (list-of-word (word-of-list2 zs))
            by right right lemma-list-of-word (word-of-list2 zs) reversed
        equals word-of-list2 xs • y • word-of-list2 zs
            by right left refl' hyp2 reversed
        equals word-of-list2 xs • word-of-list2 ys • word-of-list2 zs
            by right lemma-append2 ys zs
        equals word-of-list2 xs • (word-of-list2 (ys ++ zs))
            by lemma-append2 xs (ys ++ zs)
        equals word-of-list2 (xs ++ ys ++ zs)
            by refl' (Eq.context word-of-list2 (lemma-split3 n m us eq))
        equals word-of-list2 us

  -- The at-head tactic: apply an axiom or a lemma at the beginning of a
  -- list. See above for usage information.
  at-head : ∀ {X} {Γ : Context X} {u v k} -> Γ ⊢ u === v -> Γ ⊢ word-of-list (list-of-word u ++ k) === word-of-list (list-of-word v ++ k)
  at-head {X} {Γ} {u} {v} {k} hyp =
     equational word-of-list (list-of-word u ++ k)
             by lemma-append (list-of-word u) k reversed
         equals word-of-list (list-of-word u) • word-of-list k
             by left lemma-list-of-word u reversed
         equals u • word-of-list k
             by left hyp
         equals v • word-of-list k
             by left lemma-list-of-word v
         equals word-of-list (list-of-word v) • word-of-list k
             by lemma-append (list-of-word v) k
         equals word-of-list (list-of-word v ++ k)

  -- The in-context tactic: apply an axiom at a specified position
  -- inside a word. See above for usage information.
  in-context : ∀ {X} {Γ : Context X} {s t pre post s2 t2 s3 t3} -> (n m : Nat) ->
             let s' = list-of-word2 s
                 t' = list-of-word2 t
                 m' = m + length t' - length s'
             in (mysplit n m s' , mysplit n m' t' , list-of-word s2 , list-of-word t2) == ((pre , s2 , post) , (pre , t2 , post) , list-of-word s3 , list-of-word t3) ->
                Γ ⊢ s3 === t3 -> Γ ⊢ s === t

  in-context {X} {Γ} {s} {t} {pre} {post} {s2} {t2} {s3} {t3} n m hyp lemma =
    let s' = list-of-word2 s
        t' = list-of-word2 t
        m' = m + length t' - length s'
        hyp1 = Eq.context fst hyp
        hyp2 = Eq.context fst (Eq.context snd hyp)
        hyp3 = Eq.context fst (Eq.context snd (Eq.context snd hyp))
        hyp4 = Eq.context snd (Eq.context snd (Eq.context snd hyp))
    in
    equational s
            by refl' (lemma-list-of-word2 s) reversed
        equals word-of-list2 s'
            by lemma-mysplit n m s' hyp1 reversed
        equals word-of-list pre • s2 • word-of-list post
            by right left general-assoc hyp3
        equals word-of-list pre • s3 • word-of-list post
            by right left lemma
        equals word-of-list pre • t3 • word-of-list post
            by right left general-assoc hyp4 reversed
        equals word-of-list pre • t2 • word-of-list post
            by lemma-mysplit n m' t' hyp2
        equals word-of-list2 t'
            by refl' (lemma-list-of-word2 t)
        equals t

  -- The rewrite-in-context tactic: given a multistep rewrite rule,
  -- prove the equality of two words by applying rewriting only to the
  -- subword at the specified position. See above for usage information.
  rewrite-in-context : ∀ {X Γ s t pre post s2 t2} -> (multistep : (n : Nat) -> List X -> List X) -> (lemma-multistep : (n : Nat) -> (xs : List X) -> Γ ⊢ word-of-list xs === word-of-list (multistep n xs)) -> (n m k : Nat) ->
             let s' = list-of-word2 s
                 t' = list-of-word2 t
                 m' = m + length t' - length s'
             in (mysplit n m s' , mysplit n m' t' , multistep k (list-of-word s2)) == 
                ((pre , s2 , post) , (pre , t2 , post) , multistep k (list-of-word t2)) ->
                Γ ⊢ s === t
  rewrite-in-context {X} {Γ} {s} {t} {pre} {post} {s2} {t2} multistep lemma-multistep n m k hyp =
    let s' = list-of-word2 s
        t' = list-of-word2 t
        m' = m + length t' - length s'
        hyp1 = Eq.context fst hyp
        hyp2 = Eq.context fst (Eq.context snd hyp)
        hyp3 = Eq.context snd (Eq.context snd hyp)
    in
    equational s
            by refl' (lemma-list-of-word2 s) reversed
        equals word-of-list2 s'
            by lemma-mysplit n m s' hyp1 reversed
        equals word-of-list pre • s2 • word-of-list post
            by right left lemma-list-of-word s2
        equals word-of-list pre • word-of-list (list-of-word s2) • word-of-list post
            by right left (lemma-multistep k (list-of-word s2))
        equals word-of-list pre • word-of-list (multistep k (list-of-word s2)) • word-of-list post
            by right left refl' (Eq.context word-of-list hyp3)
        equals word-of-list pre • word-of-list (multistep k (list-of-word t2)) • word-of-list post
            by right left (lemma-multistep k (list-of-word t2)) reversed
        equals word-of-list pre • word-of-list (list-of-word t2) • word-of-list post
            by right left lemma-list-of-word t2 reversed
        equals word-of-list pre • t2 • word-of-list post
            by lemma-mysplit n m' t' hyp2
        equals word-of-list2 t'
            by refl' (lemma-list-of-word2 t)
        equals t
